import random
import numpy as np
from boonsdk import app_from_env


def read_hash_as_vectors(search=None,
                         attr='analysis.boonai-image-similarity.simhash',
                         percentage=100,
                         verbose=True):
    """This function reads a similarity hash from a Zorroa server into a numpy array.
    search: a search as returned bu archivist.AssetSearch(). If empty, an empty search will be used.
    attr: the attribute name of the hash to use.

    Returns a numpy array, a list of attribute IDs, a list of labels, and a list of legends
    """

    app = app_from_env()

    # If the provided search is None, create the empty search
    if search is None:
        search = app.assets.search({"size": 10000})

    if len(search.assets) == 0:
        print('ERROR: no assets found with the given attributes. Nothing to do.')
        return

    if verbose:
        print("About to process %d assets" % int(len(search.assets)*percentage/100.))

    random.seed(42)

    all_data = []
    assets = []
    for a in search.assets:
        if random.random()*100 < percentage:
            num_hash = []
            hash = a.get_attr(attr)
            if hash is not None:
                for char in hash:
                    num_hash.append(ord(char))
                all_data.append(num_hash)
                assets.append(a)

    x = np.asarray(all_data, dtype=np.float64)

    return x, assets



import numpy as np
from PIL import Image
import boonsdk


def download_proxy(entity, level=-1):
    """ Download image proxy from an asset or clip

    Args:
        entity: (boonsdk.entity.asset.Asset or boonsdk.entity.clip.VideoClip)
                asset or clip to get a proxy for
        level: (int) proxy image type (-1 is largest proxy)

    Returns:
        np.ndarray: img as numpy array
    """
    app = boonsdk.app_from_env()

    if type(entity) == boonsdk.entity.asset.Asset:
        proxies = entity.get_files(category="proxy",
                                   mimetype="image/",
                                   sort_func=lambda f: f.attrs["width"])

        if not proxies:
            return None

        if level >= len(proxies):
            level = -1

        proxy = proxies[level]
        img_data = app.assets.download_file(proxy.id)

    elif type(entity) == boonsdk.entity.clip.VideoClip:
        img_data = app.assets.download_file(entity.files[0])

    if not img_data:
        return None
    else:
        img = np.array(Image.open(img_data))
        return img



import pandas as pd
from collections import defaultdict

from boonsdk import app_from_env


def search_to_df(search=None,
                 num_assets=1000,
                 attrs=None,
                 descriptor='source.filename'):
    """Convert search results to DataFrame

    Args:
        search (AssetSearchResult): an AssetSearchResult instance from ES query
        num_assets (int): number of Assets to display (default: 1000)
        attrs (List[str]): attributes to get
        descriptor (str, default: source.filename): unique name to describe each row

    Returns:
        pd.DataFrame - DataFrame converted from assets
    """
    app = app_from_env()
    asset_dict = defaultdict(list)

    if search is None:
        search = app.assets.search({"size": num_assets})

    if attrs is None:
        attrs = ['media.height', 'media.width']

    for asset in search:
        src = asset.get_attr(descriptor)
        asset_dict[descriptor].append(src)
        for attr in attrs:
            a = asset.get_attr(attr)
            asset_dict[attr].append(a)
    df = pd.DataFrame(asset_dict)

    return df



import os
import cv2
import matplotlib.pyplot as plt
from IPython.display import HTML, display

from boonsdk import app_from_env
from .proxies import download_proxy

DISPLAY_HTML = "<img style='width: 200px; height: 200px; object-fit: contain; " \
               "margin: 3px; float: left; border: 2px solid black;' title='index: %d' src='%s' />"


def is_running_from_ipython():
    """Check that script is running from Jupyter notebook

    Returns:
        bool: True if in notebook otherwise False
    """
    from IPython import get_ipython
    return get_ipython() is not None


def show_thumbnails(search=None):
    """This function displays the thumbnails for the first page of assets for a given search.

    Args:
        search: Zorroa search or list of assets

    Returns:
        List[str]: list of image paths
    """
    app = app_from_env()

    if not os.path.exists('tmp'):
        os.makedirs('tmp')

    if search is None:
        search = app.assets.search({"size": 10})

    paths = []
    for entity in search:
        name = 'tmp/' + str(entity.id) + '.jpg'
        paths.append(name)
        img = download_proxy(entity, 0)
        if img is None:
            continue
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        cv2.imwrite(name, img)

    images_list = ''.join([DISPLAY_HTML % (i, str(s)) for i, s in enumerate(paths)])

    if is_running_from_ipython():
        display(HTML(images_list))

    return paths


def show_asset(asset):
    """This function downloads and displays the largest proxy for an asset

    Args:
        asset: (boonsdk.entity.asset.Asset) single analyzed processed file

    Returns:
        dict: asset's metadata
    """
    img = download_proxy(asset, -1)

    if is_running_from_ipython():
        plt.figure(figsize=(20, 10))
        plt.imshow(img)
        plt.show()

    return asset.document



#!/usr/bin/env python3
from setuptools import setup

# See https://packaging.python.org/tutorials/packaging-projects/
# for details about packaging python projects

# Generating distribution archives (run from same directory as this file)
# python3 -m pip install --user --upgrade setuptools wheel
# python3 setup.py sdist bdist_wheel

requirements = [
    'cmake',
    'boonsdk',
    'pandas',
    'matplotlib',
    'opencv-python',
    'Pillow',
    'ipython',
    'opencv_python',
    'scikit-learn',
    'bokeh',
    'holoviews',
    'MulticoreTSNE'
]

setup(
    name='boonlab',
    version='1.0.0',
    description='Boon AI Jupyter Environment',
    url='https://www.boonai.io',
    license='Apache2',
    package_dir={'': 'pylib'},
    packages=['boonlab'],
    scripts=[],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent'
    ],

    include_package_data=True,
    install_requires=requirements
)



import os
import cv2
from MulticoreTSNE import MulticoreTSNE as TSNE
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin_min

from bokeh.plotting import figure, show
from bokeh.palettes import Paired12, Spectral11
from bokeh.models import ColumnDataSource, HoverTool, PanTool, WheelZoomTool
from bokeh.models import ResetTool, SaveTool, BoxZoomTool

import holoviews as hv

from boonsdk import app_from_env
from .hashes import read_hash_as_vectors
from .proxies import download_proxy
from .display import is_running_from_ipython


def plot_tsne(search=None,
              attr='analysis.boonai-image-similarity.simhash',
              percentage=100,
              maxPoints=10000,
              nClusters=1,
              verbose=True,
              thumbs=False):
    """Plots a t-sne graph of the search provided.

    Returns an array of assets, the 2D points, and an array of cluster indices.
    """

    hv.extension('bokeh')
    app = app_from_env()

    if search is None:
        search = app.assets.search({"size": 10000})

    if len(search.assets) < 10:
        print("Less than 10 assets, returning!")
        return None, None, None

    if len(search.assets) * percentage / 100 > maxPoints:
        percentage = maxPoints * 100 / len(search.assets)

    x, assets = read_hash_as_vectors(attr=attr, search=search,
                                     percentage=percentage, verbose=verbose)

    xtsne = TSNE(n_components=2,
                 random_state=42,
                 metric='cityblock',
                 verbose=verbose).fit_transform(x)

    imgs = []

    if thumbs:
        tmpdir = 'tmp/'
        if not os.path.isdir(tmpdir):
            os.mkdir(tmpdir)
        for a in assets:
            name = 'tmp/' + a.id + '.jpg'
            fullname = './' + name
            if not os.path.isfile(fullname):
                img = download_proxy(a, 0)
                if img is None:
                    continue
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                cv2.imwrite(name, img)

            imgs.append(name)

    legend = []
    kmeans = KMeans(n_clusters=nClusters, random_state=0).fit(x)

    if is_running_from_ipython():
        closest, _ = pairwise_distances_argmin_min(kmeans.cluster_centers_, x)
        PAL = Paired12 + Spectral11
        colormap = []

        for label in kmeans.labels_:
            colormap.append(PAL[label])

        for a in assets:
            legend.append(a.get_attr('source.filename'))

        TOOLS = [BoxZoomTool, PanTool, WheelZoomTool, ResetTool, SaveTool]

        radius = 20

        if thumbs:
            source = ColumnDataSource(data=dict(
                x=xtsne[:, 0],
                y=xtsne[:, 1],
                imgs=imgs,
                desc=legend,
                colors=colormap,
                sz=[0.01 * radius] * len(colormap)
            ))

            hover = HoverTool(
                tooltips="""
                    <div>
                    <div>
                            <img
                                src=@imgs height="130" alt="@imgs" width="200"
                                style="float: left; margin: 5px 5px 5px 5px;"
                                border="2">
                    </img>
                    </div>
                    </div>
                    """
            )
        else:
            source = ColumnDataSource(data=dict(
                x=xtsne[:, 0],
                y=xtsne[:, 1],
                desc=legend,
                cluster=kmeans.labels_,
                colors=colormap,
                sz=[0.01 * radius] * len(colormap)
            ))

            hover = HoverTool(tooltips=[('Cluster', '@cluster'),
                                        ('Filename', '@desc')])

        tools = [t() for t in TOOLS] + [hover]

        p = figure(tools=tools, plot_width=800, plot_height=600)
        p.background_fill_color = "black"
        p.background_fill_alpha = 0.8
        p.grid.visible = False

        p.circle('x', 'y', fill_color='colors', radius='sz', source=source,
                 line_color=None)

        show(p)

    return assets, x, kmeans.labels_
